@@ -4,18 +4,19 @@
 import sys
 
 # Import shortcut
-from .client import RedisCluster
+from .client import StrictRedisCluster, RedisCluster
 from .pipeline import StrictClusterPipeline
 from .pubsub import ClusterPubSub
 
 # Monkey patch RedisCluster class into redis for easy access
 import redis
+setattr(redis, "StrictRedisCluster", StrictRedisCluster)
 setattr(redis, "RedisCluster", RedisCluster)
 setattr(redis, "ClusterPubSub", ClusterPubSub)
 setattr(redis, "StrictClusterPipeline", StrictClusterPipeline)
 
 # Major, Minor, Fix version
-__version__ = (0, 2, 0)
+__version__ = (1, 3, 3)
 
 if sys.version_info[0:3] == (3, 4, 0):
     raise RuntimeError("CRITICAL: rediscluster do not work with python 3.4.0. Please use 3.4.1 or higher.")
