@@ -1,76 +1,102 @@
 # -*- coding: utf-8 -*-
 
 # python std lib
-import random
 import sys
-import time
-import threading
 
 # rediscluster imports
-from .client import RedisCluster
-from .connection import by_node_context
-from .exceptions import RedisClusterException, ClusterDownException
-from .utils import clusterdown_wrapper
+from .client import StrictRedisCluster
+from .exceptions import (
+    RedisClusterException, AskError, MovedError, TryAgainError,
+)
+from .utils import clusterdown_wrapper, dict_merge
 
 # 3rd party imports
 from redis import StrictRedis
-from redis.exceptions import ResponseError, ConnectionError
-from redis._compat import imap, unicode, xrange
+from redis.exceptions import ConnectionError, RedisError, TimeoutError
+from redis._compat import imap, unicode
 
 
-class StrictClusterPipeline(RedisCluster):
+ERRORS_ALLOW_RETRY = (ConnectionError, TimeoutError, MovedError, AskError, TryAgainError)
+
+
+class StrictClusterPipeline(StrictRedisCluster):
     """
     """
 
-    def __init__(self, connection_pool, nodes_callbacks=None, result_callbacks=None, response_callbacks=None, startup_nodes=[], connections=[], opt={}, refresh_table_asap=False, slots={}, nodes=[], use_threads=True):
-        self.connection_pool = connection_pool
-        self.startup_nodes = startup_nodes
-        self.refresh_table_asap = refresh_table_asap
+    def __init__(self, connection_pool, result_callbacks=None,
+                 response_callbacks=None, startup_nodes=None):
+        """
+        """
         self.command_stack = []
-
-        self.nodes_callbacks = nodes_callbacks
-        self.result_callbacks = result_callbacks
-        self.response_callbacks = response_callbacks
-        self.use_threads = use_threads
+        self.refresh_table_asap = False
+        self.connection_pool = connection_pool
+        self.result_callbacks = result_callbacks or self.__class__.RESULT_CALLBACKS.copy()
+        self.startup_nodes = startup_nodes if startup_nodes else []
+        self.nodes_flags = self.__class__.NODES_FLAGS.copy()
+        self.response_callbacks = dict_merge(response_callbacks or self.__class__.RESPONSE_CALLBACKS.copy(),
+                                             self.CLUSTER_COMMANDS_RESPONSE_CALLBACKS)
 
     def __repr__(self):
-        return "{}".format(type(self).__name__)
+        """
+        """
+        return "{0}".format(type(self).__name__)
 
     def __enter__(self):
+        """
+        """
         return self
 
     def __exit__(self, exc_type, exc_value, traceback):
+        """
+        """
         self.reset()
 
     def __del__(self):
+        """
+        """
         self.reset()
 
     def __len__(self):
+        """
+        """
         return len(self.command_stack)
 
     def execute_command(self, *args, **kwargs):
+        """
+        """
         return self.pipeline_execute_command(*args, **kwargs)
 
     def pipeline_execute_command(self, *args, **options):
-        self.command_stack.append((args, options))
+        """
+        """
+        self.command_stack.append(PipelineCommand(args, options, len(self.command_stack)))
         return self
 
-    def raise_first_error(self, commands, response):
-        for i, r in enumerate(response):
-            if isinstance(r, ResponseError):
-                self.annotate_exception(r, i + 1, commands[i][0])
+    def raise_first_error(self, stack):
+        """
+        """
+        for c in stack:
+            r = c.result
+            if isinstance(r, Exception):
+                self.annotate_exception(r, c.position + 1, c.args)
                 raise r
 
     def annotate_exception(self, exception, number, command):
+        """
+        """
         cmd = unicode(' ').join(imap(unicode, command))
-        msg = unicode('Command # %d (%s) of pipeline caused error: %s') % (
+        msg = unicode('Command # {0} ({1}) of pipeline caused error: {2}').format(
             number, cmd, unicode(exception.args[0]))
         exception.args = (msg,) + exception.args[1:]
 
     def execute(self, raise_on_error=True):
+        """
+        """
         stack = self.command_stack
+
         if not stack:
             return []
+
         try:
             return self.send_cluster_commands(stack, raise_on_error)
         finally:
@@ -81,7 +107,6 @@ class StrictClusterPipeline(RedisCluster):
         Reset back to empty pipeline.
         """
         self.command_stack = []
-        self.refresh_table_asap = False
 
         self.scripts = set()
 
@@ -117,110 +142,94 @@ class StrictClusterPipeline(RedisCluster):
         `allow_redirections` If the pipeline should follow `ASK` & `MOVED` responses
         automatically. If set to false it will raise RedisClusterException.
         """
-        if self.refresh_table_asap:
-            self.connection_pool.nodes.initialize()
-            self.refresh_table_asap = False
+        # the first time sending the commands we send all of the commands that were queued up.
+        # if we have to run through it again, we only retry the commands that failed.
+        attempt = sorted(stack, key=lambda x: x.position)
 
-        ttl = self.RedisClusterRequestTTL
-        response = {}
-        attempt = xrange(0, len(stack)) if stack else []  # noqa
-        ask_slots = {}
+        # build a list of node objects based on node names we need to
+        nodes = {}
 
-        while attempt and ttl > 0:
-            ttl -= 1
-            node_commands = {}
-            nodes = {}
+        # as we move through each command that still needs to be processed,
+        # we figure out the slot number that command maps to, then from the slot determine the node.
+        for c in attempt:
+            # refer to our internal node -> slot table that tells us where a given
+            # command should route to.
+            slot = self._determine_slot(*c.args)
+            node = self.connection_pool.get_node_by_slot(slot)
 
-            # Keep this section so that we can determine what nodes to contact
-            for i in attempt:
-                c = stack[i]
-                slot = self._determine_slot(*c[0], **c[1])
-                if slot in ask_slots:
-                    node = ask_slots[slot]
-                else:
-                    node = self.connection_pool.nodes.slots[slot]
+            # little hack to make sure the node name is populated. probably could clean this up.
+            self.connection_pool.nodes.set_node_name(node)
 
-                self.connection_pool.nodes.set_node_name(node)
-                node_name = node['name']
-                nodes[node_name] = node
-                node_commands.setdefault(node_name, {})
-                node_commands[node_name][i] = c
+            # now that we know the name of the node ( it's just a string in the form of host:port )
+            # we can build a list of commands for each node.
+            node_name = node['name']
+            if node_name not in nodes:
+                nodes[node_name] = NodeCommands(self.parse_response, self.connection_pool.get_connection_by_node(node))
 
-            # Get one connection at a time from the pool and basiccly copy the logic from
-            #  _execute_pipeline() below and do what a normal pipeline does.
+            nodes[node_name].append(c)
 
-            attempt = []
-            if self.use_threads and len(node_commands) > 1:
-                workers = dict()
-                for node_name in node_commands:
-                    node = nodes[node_name]
-                    cmds = [node_commands[node_name][i] for i in sorted(node_commands[node_name].keys())]
-                    with by_node_context(self.connection_pool, node) as connection:
-                        workers[node_name] = Worker(self._execute_pipeline, connection, cmds, False)
-                        workers[node_name].start()
+        # send the commands in sequence.
+        # we  write to all the open sockets for each node first, before reading anything
+        # this allows us to flush all the requests out across the network essentially in parallel
+        # so that we can read them all in parallel as they come back.
+        # we dont' multiplex on the sockets as they come available, but that shouldn't make too much difference.
+        node_commands = nodes.values()
+        for n in node_commands:
+            n.write()
 
-                for node_name, worker in workers.items():
-                    worker.join()
-                    if worker.exception:
-                        for i in sorted(node_commands[node_name].keys()):
-                            response[i] = worker.exception
-                    else:
-                        for i, v in zip(sorted(node_commands[node_name].keys()), worker.value):
-                            response[i] = v
-                del workers
-            else:
-                for node_name in node_commands:
-                    node = nodes[node_name]
-                    cmds = [node_commands[node_name][i] for i in sorted(node_commands[node_name].keys())]
-                    with by_node_context(self.connection_pool, node) as connection:
-                        result = zip(
-                            sorted(node_commands[node_name].keys()),
-                            self._execute_pipeline(connection, cmds, False))
-                        for i, v in result:
-                            response[i] = v
+        for n in node_commands:
+            n.read()
 
-            ask_slots = {}
-            for i, v in response.items():
-                if not isinstance(v, Exception):
-                    continue
+        # release all of the redis connections we allocated earlier back into the connection pool.
+        # we used to do this step as part of a try/finally block, but it is really dangerous to
+        # release connections back into the pool if for some reason the socket has data still left in it
+        # from a previous operation. The write and read operations already have try/catch around them for
+        # all known types of errors including connection and socket level errors.
+        # So if we hit an exception, something really bad happened and putting any of
+        # these connections back into the pool is a very bad idea.
+        # the socket might have unread buffer still sitting in it, and then the
+        # next time we read from it we pass the buffered result back from a previous
+        # command and every single request after to that connection will always get
+        # a mismatched result. (not just theoretical, I saw this happen on production x.x).
+        for n in nodes.values():
+            self.connection_pool.release(n.connection)
 
-                if isinstance(v, ConnectionError):
-                    ask_slots[self.connection_pool.nodes.keyslot(stack[i][0][1])] = random.choice(self.startup_nodes)
-                    attempt.append(i)
-                    if ttl < self.RedisClusterRequestTTL / 2:
-                        time.sleep(0.1)
-                    continue
+        # if the response isn't an exception it is a valid response from the node
+        # we're all done with that command, YAY!
+        # if we have more commands to attempt, we've run into problems.
+        # collect all the commands we are allowed to retry.
+        # (MOVED, ASK, or connection errors or timeout errors)
+        attempt = sorted([c for c in attempt if isinstance(c.result, ERRORS_ALLOW_RETRY)], key=lambda x: x.position)
+        if attempt and allow_redirections:
+            # RETRY MAGIC HAPPENS HERE!
+            # send these remaing comamnds one at a time using `execute_command`
+            # in the main client. This keeps our retry logic in one place mostly,
+            # and allows us to be more confident in correctness of behavior.
+            # at this point any speed gains from pipelining have been lost
+            # anyway, so we might as well make the best attempt to get the correct
+            # behavior.
+            #
+            # The client command will handle retries for each individual command
+            # sequentially as we pass each one into `execute_command`. Any exceptions
+            # that bubble out should only appear once all retries have been exhausted.
+            #
+            # If a lot of commands have failed, we'll be setting the
+            # flag to rebuild the slots table from scratch. So MOVED errors should
+            # correct themselves fairly quickly.
+            self.connection_pool.nodes.increment_reinitialize_counter(len(attempt))
+            for c in attempt:
+                try:
+                    # send each command individually like we do in the main client.
+                    c.result = super(StrictClusterPipeline, self).execute_command(*c.args, **c.options)
+                except RedisError as e:
+                    c.result = e
 
-                errv = RedisCluster._exception_message(v)
-                if errv is None:
-                    continue
+        # turn the response back into a simple flat array that corresponds
+        # to the sequence of commands issued in the stack in pipeline.execute()
+        response = [c.result for c in sorted(stack, key=lambda x: x.position)]
 
-                if errv.startswith('CLUSTERDOWN'):
-                    self.connection_pool.disconnect()
-                    self.connection_pool.reset()
-                    self.refresh_table_asap = True
-                    raise ClusterDownException()
-
-                redir = self.parse_redirection_exception_msg(errv)
-
-                if not redir:
-                    continue
-
-                if redir['action'] == "MOVED":
-                    self.refresh_table_asap = True
-                    node = self.connection_pool.nodes.set_node(redir['host'], redir['port'], server_type='master')
-                    self.connection_pool.nodes.slots[redir['slot']] = node
-                    attempt.append(i)
-                    self._fail_on_redirect(allow_redirections)
-                elif redir['action'] == "ASK":
-                    node = self.connection_pool.nodes.set_node(redir['host'], redir['port'], server_type='master')
-                    ask_slots[redir['slot']] = node
-                    attempt.append(i)
-                    self._fail_on_redirect(allow_redirections)
-
-        response = [response[k] for k in sorted(response.keys())]
         if raise_on_error:
-            self.raise_first_error(stack, response)
+            self.raise_first_error(stack)
 
         return response
 
@@ -230,47 +239,39 @@ class StrictClusterPipeline(RedisCluster):
         if not allow_redirections:
             raise RedisClusterException("ASK & MOVED redirection not allowed in this pipeline")
 
-    def _execute_pipeline(self, connection, commands, raise_on_error):
-        """
-        Code borrowed from StrictRedis so it can be fixed
-        """
-        # build up all commands into a single request to increase network perf
-        all_cmds = connection.pack_commands([args for args, _ in commands])
-        try:
-            connection.send_packed_command(all_cmds)
-        except ConnectionError as e:
-            return [e for _ in xrange(len(commands))]  # noqa
-
-        response = []
-        for args, options in commands:
-            try:
-                response.append(self.parse_response(connection, args[0], **options))
-            except (ConnectionError, ResponseError):
-                response.append(sys.exc_info()[1])
-
-        if raise_on_error:
-            self.raise_first_error(commands, response)
-        return response
-
     def multi(self):
+        """
+        """
         raise RedisClusterException("method multi() is not implemented")
 
     def immediate_execute_command(self, *args, **options):
+        """
+        """
         raise RedisClusterException("method immediate_execute_command() is not implemented")
 
-    def _execute_transaction(self, connection, commands, raise_on_error):
+    def _execute_transaction(self, *args, **kwargs):
+        """
+        """
         raise RedisClusterException("method _execute_transaction() is not implemented")
 
     def load_scripts(self):
+        """
+        """
         raise RedisClusterException("method load_scripts() is not implemented")
 
     def watch(self, *names):
+        """
+        """
         raise RedisClusterException("method watch() is not implemented")
 
     def unwatch(self):
+        """
+        """
         raise RedisClusterException("method unwatch() is not implemented")
 
-    def script_load_for_pipeline(self, script):
+    def script_load_for_pipeline(self, *args, **kwargs):
+        """
+        """
         raise RedisClusterException("method script_load_for_pipeline() is not implemented")
 
     def delete(self, *names):
@@ -288,7 +289,8 @@ def block_pipeline_command(func):
     Prints error because some pipelined commands should be blocked when running in cluster-mode
     """
     def inner(*args, **kwargs):
-        raise RedisClusterException("ERROR: Calling pipelined function {} is blocked when running redis in cluster mode...".format(func.__name__))
+        raise RedisClusterException("ERROR: Calling pipelined function {0} is blocked when running redis in cluster mode...".format(func.__name__))
+
     return inner
 
 
@@ -318,6 +320,7 @@ StrictClusterPipeline.move = block_pipeline_command(StrictRedis.move)
 StrictClusterPipeline.mset = block_pipeline_command(StrictRedis.mset)
 StrictClusterPipeline.msetnx = block_pipeline_command(StrictRedis.msetnx)
 StrictClusterPipeline.pfmerge = block_pipeline_command(StrictRedis.pfmerge)
+StrictClusterPipeline.pfcount = block_pipeline_command(StrictRedis.pfcount)
 StrictClusterPipeline.ping = block_pipeline_command(StrictRedis.ping)
 StrictClusterPipeline.publish = block_pipeline_command(StrictRedis.publish)
 StrictClusterPipeline.randomkey = block_pipeline_command(StrictRedis.randomkey)
@@ -354,23 +357,79 @@ StrictClusterPipeline.sunionstore = block_pipeline_command(StrictRedis.sunionsto
 StrictClusterPipeline.time = block_pipeline_command(StrictRedis.time)
 
 
-class Worker(threading.Thread):
+class PipelineCommand(object):
     """
-    A simple thread wrapper class to perform an arbitrary function and store the result.
-    Used to parallelize network i/o when issuing many pipelined commands to multiple nodes in the cluster.
-    Later on we may need to convert this to a thread pool, although if you use gevent the current implementation
-    is not too bad.
     """
-    def __init__(self, func, *args, **kwargs):
-        threading.Thread.__init__(self)
-        self.func = func
-        self.args = args
-        self.kwargs = kwargs
-        self.exception = None
-        self.value = None
 
-    def run(self):
+    def __init__(self, args, options=None, position=None):
+        self.args = args
+        if options is None:
+            options = {}
+        self.options = options
+        self.position = position
+        self.result = None
+        self.node = None
+        self.asking = False
+
+
+class NodeCommands(object):
+    """
+    """
+
+    def __init__(self, parse_response, connection):
+        """
+        """
+        self.parse_response = parse_response
+        self.connection = connection
+        self.commands = []
+
+    def append(self, c):
+        """
+        """
+        self.commands.append(c)
+
+    def write(self):
+        """
+        Code borrowed from StrictRedis so it can be fixed
+        """
+        connection = self.connection
+        commands = self.commands
+
+        # We are going to clobber the commands with the write, so go ahead
+        # and ensure that nothing is sitting there from a previous run.
+        for c in commands:
+            c.result = None
+
+        # build up all commands into a single request to increase network perf
+        # send all the commands and catch connection and timeout errors.
         try:
-            self.value = self.func(*self.args, **self.kwargs)
-        except Exception as e:
-            self.exception = e
+            connection.send_packed_command(connection.pack_commands([c.args for c in commands]))
+        except (ConnectionError, TimeoutError) as e:
+            for c in commands:
+                c.result = e
+
+    def read(self):
+        """
+        """
+        connection = self.connection
+        for c in self.commands:
+
+            # if there is a result on this command, it means we ran into an exception
+            # like a connection error. Trying to parse a response on a connection that
+            # is no longer open will result in a connection error raised by redis-py.
+            # but redis-py doesn't check in parse_response that the sock object is
+            # still set and if you try to read from a closed connection, it will
+            # result in an AttributeError because it will do a readline() call on None.
+            # This can have all kinds of nasty side-effects.
+            # Treating this case as a connection error is fine because it will dump
+            # the connection object back into the pool and on the next write, it will
+            # explicitly open the connection and all will be well.
+            if c.result is None:
+                try:
+                    c.result = self.parse_response(connection, c.args[0], **c.options)
+                except (ConnectionError, TimeoutError) as e:
+                    for c in self.commands:
+                        c.result = e
+                        return
+                except RedisError:
+                    c.result = sys.exc_info()[1]
